#include "miniaudioengine.h"
#include <iostream> //enum macro
MiniAudioEngine::MiniAudioEngine()
{
    for (int i =0; i < MAX_LAYERS; i++) {
        m_mediaLoaded[i] = false;
        m_currentLayerValues[i].status = Status::Iddle;
        m_currentLayerValues[i].pan = 128;
        m_currentLayerValues[i].pitch = 128;
        m_currentLayerValues[i].vol = 0;
        m_currentLayerValues[i].cursor = 0;
    }
}

void MiniAudioEngine::audioDataCallback(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount)
{
    (void)pInput;
    //Do master audio processing before sending to device.
    ma_engine_read_pcm_frames((ma_engine*)pDevice->pUserData, pOutput, frameCount, NULL);
}

void MiniAudioEngine::stopEngine()
{
    ma_engine_uninit(&engine[0]);
    ma_device_uninit(&device[0]);
    ma_context_uninit(&context);
    ma_resource_manager_uninit(&resourceManager);
}

bool MiniAudioEngine::startEngine()
{
    ma_result result;

    result = this->startContext();
    if (result != MA_SUCCESS) return result;
    result = this->getAllAudioDevices();
    return result;
}

ma_result MiniAudioEngine::startDevice(uint systemId, uint internalId)
{
    ma_result result;
    ma_device_config deviceConfig;
    ma_engine_config engineConfig;

    if (systemId >= playbackDeviceCount)
        systemId = playbackDeviceCount - 1;
    deviceConfig = ma_device_config_init(ma_device_type_playback);
    deviceConfig.playback.pDeviceID = &pPlaybackDeviceInfos[systemId].id;
    deviceConfig.playback.format    = resourceManager.config.decodedFormat;
    deviceConfig.playback.channels  = 0;
    deviceConfig.sampleRate         = resourceManager.config.decodedSampleRate;
    deviceConfig.dataCallback       = audioDataCallback;
    deviceConfig.pUserData          = &engine[internalId];
    result = ma_device_init(&context, &deviceConfig, &device[internalId]);
    if (result != MA_SUCCESS) {
        qCritical("Failed to initialize audio device %s.", pPlaybackDeviceInfos[systemId].name);
        return result;
    }
    engineConfig = ma_engine_config_init();
    engineConfig.pDevice          = &device[internalId];
    engineConfig.pResourceManager = &resourceManager;
    engineConfig.noAutoStart      = MA_TRUE;
    result = ma_engine_init(NULL, &engine[internalId]);
    if (result != MA_SUCCESS) {
        qCritical("Failed to initialize audio engine.");
        return result;
    }
    result = ma_engine_start(&engine[internalId]);
    if (result != MA_SUCCESS) {
        qCritical("Failed to start audio engine %i.", systemId);
        return result;
    }
    qInfo("Initialized audio device internal: %ui system: %d %s", internalId, systemId, pPlaybackDeviceInfos[systemId].name);
    return result;
}

ma_result MiniAudioEngine::startContext()
{
    ma_result result;

    resourceManagerConfig = ma_resource_manager_config_init();
    resourceManagerConfig.decodedFormat     = ma_format_f32;    /* ma_format_f32 should almost always be used as that's what the engine (and most everything else) uses for mixing. */
    resourceManagerConfig.decodedChannels   = 0;
    resourceManagerConfig.decodedSampleRate = ma_standard_sample_rate_44100;
    resourceManagerConfig.jobThreadCount    = 4;
    result = ma_resource_manager_init(&resourceManagerConfig, &resourceManager);
    if (result != MA_SUCCESS) {
        qCritical("Failed to initialize audio resource manager.");
        return result;
    }
    result = ma_context_init(NULL, 0, NULL, &context);
    if (result != MA_SUCCESS) {
        qCritical("Failed to initialize audio context.");
    }
    return result;
}

// enum all audio devices in system
ma_result MiniAudioEngine::getAllAudioDevices()
{
    ma_result result;

    result = ma_context_get_devices(&context, &pPlaybackDeviceInfos, &playbackDeviceCount, NULL, NULL);
    if (result != MA_SUCCESS) {
        qWarning("Failed to enumerate playback devices.\n");
        ma_context_uninit(&context);
        return result;
    }
    printf("Audio devices available:\n");
    for (ma_uint32 iAvailableDevice = 0; iAvailableDevice < playbackDeviceCount; iAvailableDevice += 1) {
        qInfo("%d: : %s", iAvailableDevice, pPlaybackDeviceInfos[iAvailableDevice].name);
    }
    return result;
}

ma_result MiniAudioEngine::loadMedia(int layer, char *file, uint audioDevice)
{
    ma_result result;

    if (m_mediaLoaded[layer] == true)
    {
        ma_sound_uninit(&m_currentSound[layer]);
        m_mediaLoaded[layer] = false;
    }
    result = ma_sound_init_from_file(&engine[audioDevice], file, \
                MA_SOUND_FLAG_NO_SPATIALIZATION \
              | MA_SOUND_FLAG_DECODE \
              | MA_SOUND_FLAG_STREAM \
            /*| MA_SOUND_FLAG_NO_PITCH \*/
                , NULL, NULL, &m_currentSound[layer]);
    if (result != MA_SUCCESS)
        qWarning("Failed to load file %s", file);
    else {
        m_mediaLoaded[layer] = true;
        this->refreshValues(layer);
        m_currentLayerValues[layer].media = file;
    }
    return result;
}

float MiniAudioEngine::getDuration(int layer)
{
    ma_result result;
    float ret;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_length_in_seconds(&m_currentSound[layer], &ret);
    if (result != MA_SUCCESS) {
        return result;
    }
    return (ret * 1000);
}

float MiniAudioEngine::getCursor(int layer)
{
    ma_result result;
    float ret = 0;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_cursor_in_seconds(&m_currentSound[layer], &ret);
    if (result != MA_SUCCESS)
    {
        qWarning("%i can not get cursor error %i", layer, result);
        ret = MA_ERROR;
    }
    return ret;
}

ma_result MiniAudioEngine::printFormatInfo(int layer)
{
    ma_format format;
    ma_uint32 channels;
    ma_uint32 sampleRate;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    ma_result result = ma_sound_get_data_format(&m_currentSound[layer], &format, &channels, &sampleRate, NULL, 0);
    if (result != MA_SUCCESS)
        qWarning("%i failed to get data format %i\n", layer, result);
    else
        qInfo() << "Layer:" << layer << m_currentLayerValues[layer].media << "samples/sec:" << sampleRate << "format:" <<  format << "channels:" << channels;
    return result;
}

// Expects between 0 and 1 vol value
void MiniAudioEngine::volChanged(int layer, float vol)
{
    if (m_mediaLoaded[layer] == false)
        return;
    if (vol >= 1)
        vol = 0.99f;
    ma_sound_group_set_fade_in_milliseconds(&m_currentSound[layer], -1, vol, FADE_TIME);
    m_currentLayerValues[layer].vol = vol;
}

void MiniAudioEngine::panChanged(int layer, float value)
{
    float result;

    if (m_mediaLoaded[layer] == false)
        return;
    result = (value / 128.0) - 1.0;
    ma_sound_group_set_pan(&m_currentSound[layer], result);
    m_currentLayerValues[layer].pan = value;
}

void MiniAudioEngine::pitchChanged(int layer, float value)
{
    float pitch;

    if (m_mediaLoaded[layer] == false)
        return;
    pitch = value / 128.0;
    ma_sound_group_set_pitch(&m_currentSound[layer], pitch);
    m_currentLayerValues[layer].pitch = value;
}

ma_result MiniAudioEngine::playbackChanged(int layer, Status status)
{
    ma_result result = MA_SUCCESS;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    bool loop = false;
    switch (status) {
        case Status::Paused:
            result = ma_sound_stop_with_fade_in_milliseconds(&m_currentSound[layer], FADE_TIME);
            break;
        case Status::Stopped:
            ma_sound_stop_with_fade_in_milliseconds(&m_currentSound[layer], FADE_TIME);
            result = this->seekToCursor(layer, m_currentLayerValues[layer].cursor);
            break;
        case Status::PlayingLoop:
            loop = true;
        case Status::PlayingOnce:
        case Status::PlayingFolder:
        case Status::PlayingFolderLoop:
        case Status::PlayingFolderRandom:
            ma_sound_set_stop_time_in_milliseconds(&m_currentSound[layer], ~(ma_uint64)0);
            ma_sound_set_looping(&m_currentSound[layer], loop);
            result = ma_sound_start(&m_currentSound[layer]);
            //this->volChanged(layer, m_currentLayerValues[layer].vol); // glitch when seek to cursor, how flush the audio buffer?
        default:
            break;
    }
    if (result == MA_SUCCESS)
        m_currentLayerValues[layer].status = status;
    return result;
}

ma_result MiniAudioEngine::seekToCursor(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;
    ma_uint64 end, start;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_length_in_pcm_frames(&m_currentSound[layer], &end);
    if (result != MA_SUCCESS) { return result; }
    start = (cursor * end) / 65025;
    result = ma_sound_seek_to_pcm_frame(&m_currentSound[layer], start);
    //result = ma_data_source_set_loop_point_in_pcm_frames(&m_currentSound[layer], start, end); // this do nothing here, it must be done after set_looping or start?
    return (result);
}

ma_result MiniAudioEngine::setCursor(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;

    m_currentLayerValues[layer].cursor = cursor;
    result = this->seekToCursor(layer, cursor);
    return (result);
}

Status MiniAudioEngine::getStatus(int layer)
{
    return m_currentLayerValues[layer].status;
}

void MiniAudioEngine::refreshValues(int layer)
{
    this->seekToCursor(layer, m_currentLayerValues[layer].cursor);
    this->panChanged(layer, m_currentLayerValues[layer].pan);
    this->volChanged(layer, m_currentLayerValues[layer].vol);
    this->pitchChanged(layer, m_currentLayerValues[layer].pitch);
    this->playbackChanged(layer, m_currentLayerValues[layer].status);
}
